/* Name: asp162bl.c
 * Project: AVR USB driver for AT90USB
 * Author: Koji Suzuki
 * Creation Date: 2009-12-16
 * Original Project: USBaspLoader
 * Original Author: Christian Starkjohann
 * Original Creation Date: 2007-12-08
 * Tabsize: 8
 * Copyright: (c) 2007 by OBJECTIVE DEVELOPMENT Software GmbH
 * Copyright: (c) 2009 by Koji Suzuki (suz@nmj2.sakura.ne.jp)
 * License: GNU GPL v2 (see License.txt)
 * Original Revision: Id: main.c 730 2009-03-20 09:05:11Z cs 
 */

#include <string.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <avr/wdt.h>
#include <util/delay.h>
#include <usbconfig.h>
#include <usb162.h>

#include <avr/boot.h>

#define bit_set(port,bit)	{ port |= (1<<(bit));}
#define bit_clear(port,bit)	{ port &= ~(1<<(bit));}
#define min(a,b)		((a)<(b) ? (a):(b))

#define nop()	__asm__ volatile ("nop" )
#define	wdr()	__asm__ volatile("wdr" )

static void leaveBootloader() __attribute__((__noreturn__));

#include <bootloaderconfig.h>


#define NO_INTERRPT
#ifdef NO_INTERRPT
#define __cli()
#define __sei()
#else
#define __cli()	__asm__ volatile ("cli" )
#define __sei()	__asm__ volatile ("sei" )
#endif

#ifndef APP_FLASHEND
#define APP_FLASHEND	(BOOTLOADER_ADDRESS -1)
#endif

/* ------------------------------------------------------------------------ */

/* Request constants used by USBasp */
#define USBASP_FUNC_CONNECT         1
#define USBASP_FUNC_DISCONNECT      2
#define USBASP_FUNC_TRANSMIT        3
#define USBASP_FUNC_READFLASH       4
#define USBASP_FUNC_ENABLEPROG      5
#define USBASP_FUNC_WRITEFLASH      6
#define USBASP_FUNC_READEEPROM      7
#define USBASP_FUNC_WRITEEEPROM     8
#define USBASP_FUNC_SETLONGADDRESS  9
#define USBASP_FUNC_SETISPSCK       10

#define ISPCMD_CHIP_ERASE	    0xac
#define ISPCMD_READ_SIGNATURE	    0x30
#define ISPCMD_READ_EEPROM	    0xa0
#define ISPCMD_WRITE_EEPROM	    0xc0

#define ISPCMD_READ_FLASH_L	    0x20
#define ISPCMD_READ_FLASH_H	    0x28

#define ISPCMD_READ_FUSE_L	    0x50	/* lower/ext */
#define ISPCMD_READ_FUSE_H	    0x58	/* upper/lockbit */

/* Block mode flags */
#define PROG_BLOCKFLAG_FIRST    1
#define PROG_BLOCKFLAG_LAST     2

/* Block mode data size */
#define USBASP_READBLOCKSIZE   200
#define USBASP_WRITEBLOCKSIZE  200

/* ------------------------------------------------------------------------ */

/* defaults if not in config file: */
#ifndef HAVE_EEPROM_PAGED_ACCESS
#   define HAVE_EEPROM_PAGED_ACCESS 0
#endif
#ifndef HAVE_EEPROM_BYTE_ACCESS
#   define HAVE_EEPROM_BYTE_ACCESS  0
#endif
#ifndef BOOTLOADER_CAN_EXIT
#   define  BOOTLOADER_CAN_EXIT     0
#endif

/* device compatibility: */
#ifndef GICR    /* ATMega*8 don't have GICR, use MCUCR instead */
#   define GICR     MCUCR
#endif

#ifndef HAVE_LOCK_WRITE
#   define HAVE_LOCK_WRITE 0
#endif
/* ------------------------------------------------------------------------ */
#if (FLASHEND > USHRT_MAX)
#   define CURRENT_ADDRESS  currentAddress.l
#   define addr_t           uint32_t
#else
#   define CURRENT_ADDRESS  currentAddress.w[0]
#   define addr_t           uint16_t
#endif

typedef union longConverter{
    addr_t  l;
    uint16_t  w[sizeof(addr_t)/2];
    uint8_t   b[sizeof(addr_t)];
}longConverter_t;

#if BOOTLOADER_CAN_EXIT
static uint8_t            requestBootLoaderExit;
#endif
static longConverter_t  currentAddress; /* in bytes */
static uint8_t            bytesRemaining;
static uint8_t            isLastPage;
#if HAVE_EEPROM_PAGED_ACCESS
static uint8_t            currentRequest;
#endif

#if !HAVE_SIGNATURE_ACCESS
static uint8_t  signatureBytes[4] PROGMEM = {
    SIGNATURE_0, SIGNATURE_1, SIGNATURE_2, 0
};
#endif

/* ------------------------------------------------------------------------ */


static void leaveBootloader()
{
    void (*nullVector)(void) __attribute__((__noreturn__)) = NULL;
    bootLoaderExit();
    cli();

#if (BOOTLOADER_ADDRESS != 0)
    GICR = (1 << IVCE);     /* enable change of interrupt vectors */
    GICR = (0 << IVSEL);    /* move interrupts to application flash section */
#endif

/* We must go through a global function pointer variable instead of writing
 *  ((void (*)(void))0)();
 * because the compiler optimizes a constant 0 to "rcall 0" which is not
 * handled correctly by the assembler.
 */
    nullVector();
}

/* ------------------------------------------------------------------------ */
#if HAVE_FUSE_ACCESS
static inline uint8_t read_fuse(uint16_t addr) {
	SPMCSR |= (1<<SPMEN) | (1<<BLBSET);
	return __LPM(addr);
}
#endif

#if HAVE_SIGNATURE_ACCESS
static inline uint8_t read_signature(uint16_t addr) {
	SPMCSR |= (1<<SPMEN) | (1<<SIGRD);
	return  __LPM(addr);
}
#endif

#ifdef HAVE_LOCK_WRITE
static inline void lock_bits_set(uint8_t lock_bits) {
    __asm__ __volatile__				\
    (							\
        "ldi r30, 1\n\t"				\
        "ldi r31, 0\n\t"				\
        "mov r0, %2\n\t"				\
        "sts %0, %1\n\t"				\
        "spm\n\t"					\
        :						\
        : "i" (_SFR_MEM_ADDR(SPMCSR)),			\
          "r" ((uint8_t)((1<<SPMEN)|(1<<BLBSET))),	\
          "r" (lock_bits)				\
        : "r0", "r30", "r31"				\
    );							\
}
#endif

uint8_t asp_setup(uint8_t *in, uint8_t *out)
{
    uint8_t  len = 0;
    uint8_t  request = in[USB_RQ_REQ]; 
    uint8_t  value_l = in[USB_RQ_VALUE_L];
#define  ispcmd	value_l
    uint8_t  value_h = in[USB_RQ_VALUE_H];
    uint8_t  rval = 0;
    uint16_t address;

    if(request == USBASP_FUNC_TRANSMIT){   /* emulate parts of ISP protocol */
        address = (uint16_t)(value_h <<8) | in[USB_RQ_INDEX_L];
        if(ispcmd == ISPCMD_READ_SIGNATURE){
            rval = in[USB_RQ_INDEX_L] & 3;
#if HAVE_SIGNATURE_ACCESS
	    if (rval == 3) {
		    rval = 1;
	    } else {
		    rval <<= 1;
	    }
            __cli();
  #if 1
            rval = read_signature(rval);
  #else
	    SPMCSR |= (1<<SPMEN) | (1<<SIGRD);
	    rval =  __LPM(rval);
  #endif
	    __sei();
#else
            rval = __LPM(signatureBytes+rval);
#endif
#if HAVE_EEPROM_BYTE_ACCESS
        }else if(ispcmd == ISPCMD_READ_EEPROM){
            rval = eeprom_read_byte((void *)address);
        }else if(ispcmd == ISPCMD_WRITE_EEPROM){
            eeprom_write_byte((void *)address, in[USB_RQ_INDEX_H]);
	    eeprom_busy_wait();
#endif
#if HAVE_FLASH_BYTE_ACCESS
        }else if((ispcmd & 0xf7) == ISPCMD_READ_FLASH_L) {
            address <<= 1;
	    if (ispcmd & 0x08) {  /* read FLASH upper */
		    address += 1;
	    }
	    {
            	rval = __LPM((void *)address);
	    }
#endif
#if HAVE_FUSE_ACCESS
        }else if((ispcmd & 0xf7) == ISPCMD_READ_FUSE_L) {
	    rval = 0;
	    if (ispcmd & 0x08) rval |= 1;
	    if (value_h & 0x08) rval |= 2;
            __cli();
	    rval = read_fuse(rval);
            __sei();
#endif
#if HAVE_CHIP_ERASE
        }else if(ispcmd == ISPCMD_CHIP_ERASE) {
          if(value_h == 0x80){
            addr_t addr;
            for(addr = 0; addr < (APP_FLASHEND) + 1; addr += SPM_PAGESIZE) {
                /* erase page and wait */
                __cli();
                boot_page_erase(addr);
                __sei();
                boot_spm_busy_wait();
            }
#if HAVE_LOCK_WRITE
          }else if(value_h == 0xe0) {
		value_h = in[USB_RQ_INDEX_H];
#ifdef LOCK_WRITE_IGNORE_BLB01
		bit_set(value_h, 0x2); // ignore BLB01 (bootloader can write app)
#endif
#ifdef LOCK_WRITE_IGNORE_BLB02
		bit_set(value_h, 0x3); // ignore BLB02 (bootloader can read app)
#endif
#ifdef LOCK_WRITE_IGNORE_BLB12
		bit_set(value_h, 0x5); // ignore BLB12 (bootloader can read itself)
#endif
                __cli();
                lock_bits_set(value_h);
                __sei();
#endif
          }
#endif
        }else{
            /* ignore all others, return default value == 0 */
        }
        out[3] = rval;
        len = 4;
    }else if(request == USBASP_FUNC_ENABLEPROG || request == USBASP_FUNC_SETISPSCK){
        out[0] = 0;
        len = 1;
    }else if(request == USBASP_FUNC_CONNECT){
#if BOOTLOADER_CAN_EXIT
#else
	connect_led_on();
#endif
    }else if(request == USBASP_FUNC_DISCONNECT){
#if BOOTLOADER_CAN_EXIT
        requestBootLoaderExit = 1;      /* allow proper shutdown/close of connection */
#else
	connect_led_off();
#endif
    }else if(request == USBASP_FUNC_SETLONGADDRESS){
#if (FLASHEND > USHRT_MAX)
            currentAddress.b[2] = in[USB_RQ_INDEX_L];
            currentAddress.b[3] = in[USB_RQ_INDEX_H];
#endif
    }else if(request > USBASP_FUNC_SETISPSCK) {
    }else {
        currentAddress.b[0] = value_l;
        currentAddress.b[1] = value_h;
	len = in[USB_RQ_LENGTH_L];
	if(request == USBASP_FUNC_READEEPROM || request == USBASP_FUNC_READFLASH) {
    		//for(rval = len; rval ; rval--){
    		for(rval = 0; rval < len; rval++) {
		    if (request == USBASP_FUNC_READEEPROM) {
            		*out++ = eeprom_read_byte((void *)currentAddress.w[0]);
		    } else {
            		    *out++ = pgm_read_byte((void *)CURRENT_ADDRESS);
		    }
        	    CURRENT_ADDRESS++;
		}
	} else {
            /* if(in[USB_RQ_REQ] == USBASP_FUNC_WRITEFLASH) only evaluated during writeFlash anyway */
            isLastPage = in[USB_RQ_INDEX_H] & PROG_BLOCKFLAG_LAST;
#if HAVE_EEPROM_PAGED_ACCESS
            currentRequest = request;
#endif
	    bytesRemaining = len;
            len = 0; // NO ERROR @ OUT
	}
    }
    busy_led_off();
    return len;
#undef ispcmd
}

uint8_t asp_out(uint8_t *data, uint8_t len)
{
    uint8_t isLast;
    uint8_t i;

    busy_led_on();
    if(len > bytesRemaining)
        len = bytesRemaining;
    bytesRemaining -= len;
    isLast = (bytesRemaining == 0);
#if HAVE_EEPROM_PAGED_ACCESS
    if(currentRequest == USBASP_FUNC_WRITEEEPROM){
        for(i = 0; i < len; i++){
            eeprom_write_byte((void *)(currentAddress.w[0]++), *data++);
        }
	eeprom_busy_wait();
    }else{ // USBASP_FUNC_WRITEFLASH
#endif
        for(i = 0; i < len;){
#if !HAVE_CHIP_ERASE
            if((currentAddress.w[0] & (SPM_PAGESIZE - 1)) == 0){    /* if page start: erase */

                __cli();
                boot_page_erase(CURRENT_ADDRESS);   /* erase page */
                __sei();
                boot_spm_busy_wait();               /* wait until page is erased */
            }
#endif // CHIP_ERASE
            i += 2;
            __cli();
            boot_page_fill(CURRENT_ADDRESS, *(uint16_t *)data);
            __sei();
            CURRENT_ADDRESS += 2;
            data += 2;
            /* write page when we cross page boundary or we have the last partial page */
            if((currentAddress.w[0] & (SPM_PAGESIZE - 1)) == 0 
			|| (isLast && (i >= len) && isLastPage)){

                __cli();
                boot_page_write(CURRENT_ADDRESS - 2);
                __sei();
                boot_spm_busy_wait();
                __cli();
                boot_rww_enable();
                __sei();
            }
        }
#if HAVE_EEPROM_PAGED_ACCESS
    }
#endif
    busy_led_off();
    return isLast?(uint8_t)255:0; 
}
/* ------------------------------------------------------------------------ */

static inline void asp_init() {
	uint8_t i = 20;	// abount 20ms 

	busy_led_off();
	while (--i) {
		_delay_ms(1);
	}

#if 0
	usbDeviceDisconnect();
	while (--i) {
		_delay_ms(1); // abount 250ms 
	}
#endif
	usb_vendor_setup = asp_setup;
	usb_user_out = asp_out;
	busy_led_on();
	usbInit();
	sei();
	busy_led_off();
#if BOOTLOADER_CAN_EXIT
	connect_led_on();
#endif
}

int main(void)
{ 
#if BOOTLOADER_CAN_EXIT
    uint8_t i = 0, j = 0;
#endif
#ifdef F_SCALER
    uint8_t save_clkpr;
#endif

    wdt_reset();
    cli();
    bootLoaderInit();

#ifdef F_SCALER
    save_clkpr = CLKPR;
    /* set system_clock pre scaler 1/8 => 1/1 */
    CLKPR = (1<<CLKPCE);
    CLKPR = F_SCALER;
#endif

    //bit_set(REGCR, REGDIS); // disable internal regulator

    /* initialize  */
    wdt_disable();      /* main app may have enabled watchdog */

#if (BOOTLOADER_ADDRESS != 0)
    GICR = (1 << IVCE);  /* enable change of interrupt vectors */
    GICR = (1 << IVSEL); /* move interrupts to boot flash section */
#endif

    if(bootLoaderEnterCondition()){
    	bootLoaderEnter();
        asp_init();
        do{
	    wdt_reset();
            usbPoll();
#if BOOTLOADER_CAN_EXIT
            if(requestBootLoaderExit){
                if(--i == 0){
                    if(--j == 0)
                        break;
                }
            }
#endif
        }while(!bootLoaderExitCondition());  /* main event loop */
    }
    usbDeviceDisconnect();
#ifdef F_SCALER
    CLKPR = (1<<CLKPCE);
    CLKPR = save_clkpr;
#endif
    leaveBootloader();
    return 0;
}
