#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>
#include <util/delay.h>

#define bit_set(port,bit)	{ port |= (1<<(bit));}
#define bit_clear(port,bit)	{ port &= ~(1<<(bit));}
#define min(a,b)		((a)<(b) ? (a):(b))

#include <serialconfig.h>

#if defined(__CDC162__) 
#include <usbconfig.h>
#include <cdc162.h>
#else
#include <usbcdc.h>
#endif
#include <serjtag.h>

#if ( TX_SIZE < (32) )
#error "TX_SIZE is too small"
#endif

#define JTAG_BUFSIZE	64

#define can_getc	usbcdc_can_getc
#define getc		usbcdc_getc
#define putc		usbcdc_putc
#define poll		usbcdc_poll

#define swap(v) __asm__ volatile ("swap %0" : "=r" (v) : "0" (v) )
#define nop()   __asm__ volatile ("nop" )

static uint8_t extmode = 0;
uint8_t ext_buf[JTAG_BUFSIZE+1];

static inline void init_spi() {
	TGT_DDR &=  ~((1<<TGT_MISO));
	TGT_PORT |=  ((1<<TGT_RESET));
	TGT_PORT &=  ~((1<<TGT_SCK) | (1<<TGT_MOSI) |(1<<TGT_MISO) );
	TGT_DDR |=    ((1<<TGT_SCK) | (1<<TGT_MOSI) |(1<<TGT_RESET) );
#ifdef USE_USI_SPI
	USICR = (1<<USIWM0); // enable USI as SPI Master
#elif defined(USE_HW_SPI)
	/* default 250 kHz */
  #if (F_CPU == 8000000UL)
	SPCR =  (1 << SPE) | (1 << MSTR) | (2 << SPR0);
	SPSR =  (1 << SPI2X);
  #else
	SPCR =  (1 << SPE) | (1 << MSTR) | (2 << SPR0);
  #endif
#endif
}

static inline void term_spi() {
#ifdef USE_USI_SPI
	USICR = 0; // disable USI
#elif defined(USE_HW_SPI)
	SPCR =0;
	SPSR =0;
#endif
	TGT_DDR &=   ~((1<<TGT_SCK) | (1<<TGT_MOSI) | (1<<TGT_RESET) |(1<<TGT_MISO));
	TGT_PORT &=  ~((1<<TGT_SCK) | (1<<TGT_MOSI) | (1<<TGT_RESET) |(1<<TGT_MISO));
	cli();
	CLKPR = (1<<CLKPCE);
#ifdef F_SCALER
	CLKPR = F_SCALER;
#else
	CLKPR = 0;
#endif
	sei();
}

#define		par_putcx_start()
#define		par_putcx	putc
#define		par_putcx_stop()
#define		par_getcx	getc
/* JTAG COMMAND format
 *
 *  +-----------+-----+-----+-----------+-----------+    +-----------+
 *  | COMMONAD  |FLAGS|BITS |   BYTES   |   DATA 1  |... |   DATA N  |
 *  +-----------+-----+-----+-----------+-----------+    +-----------+
 *       8         5     3       8
 *
 *  N : BYTES + (BITS != 0)?1:0 
 *  DATA: MSB first bit-stream
 *  DATA length : 8 * BYTES + BITS
 *
 */
#define TGT_TDI		TGT_MOSI
#define TGT_TDO		TGT_MISO
#define TGT_TCK		TGT_SCK
#define TGT_TMS		TGT_RESET

/* SET command parameter */
#define JTAG_SET_TDI	0x80
#define JTAG_SET_TMS	0x40
#define JTAG_SET_TCK	0x20
#define JTAG_SET_DELAY	0x0f


static uint8_t clkpr;
static uint8_t jtag_delay = 1;
static uint8_t jtag_flags;
#define JTAG_RECIEVE	0x80
#define JTAG_TMS_HIGH	0x40
#define JTAG_TDI_HIGH	0x20
#define JTAG_USE_DELAY	0x10
#define JTAG_BITS	0x07
#define JTAG_CONST_TMS	0x01 /* interval */

#define jtag_buf ext_buf
static uint8_t jtag_bytes;
static uint8_t jtag_bits;
static uint8_t jtag_bitoff;

#ifdef USE_HW_SPI
static uint8_t spcr;
static uint8_t spsr;
#endif

#if 1
/* 4 clock version */
static inline void __delay_clk(uint8_t __count) {
        __asm__ volatile (
                "1: dec %0" "\n\t"
                "nop" "\n\t"
                "brne 1b"
                : "=r" (__count)
                : "0" (__count)
        );
}
#else
#define __delay_clk(v) 	
#endif

static inline void jtag_set_speed(uint8_t option) {
	if (option == SERJTAG_SPEED_AUTO)
		option = SERJTAG_SPEED_250;
#if F_CPU == 8000000UL
	if (option < SERJTAG_SPEED_4000)
		option = SERJTAG_SPEED_4000;
#endif
#ifdef F_SCALER
	clkpr = F_SCALER;
#else
	clkpr = 0;
#endif
	while (option > SERJTAG_SPEED_62_5) {
		--option;
		clkpr ++;
	}
	jtag_delay = (uint8_t)(F_CPU/4000000UL) << (option -3);
	             /* 2 @8Mhz, 3@12Mhz */
	if (jtag_delay > 2) jtag_delay -= 2;
	if (jtag_delay == 0) jtag_delay = 1; // 0 means 0x100

#ifdef USE_HW_SPI
	bit_clear(spsr, SPI2X);
  #if F_CPU == 8000000UL
	if (!(option & 1)) bit_set(spsr, SPI2X);
	option = (option -2) >> 1;
	spcr = (option << SPR0) | (1 << SPE) | (1 << MSTR);
  #elif F_CPU == 16000000UL
	if (option & 1) bit_set(spsr, SPI2X);
	option = (option -1) >> 1;
	spcr = (option << SPR0) | (1 << SPE) | (1 << MSTR);
  #endif
#endif
}

static inline void jtag_set(uint8_t flags) {
	jtag_set_speed(flags & JTAG_SET_DELAY);
	if (flags & JTAG_SET_TDI) {
#ifdef USE_USI_SPI
		bit_set(USIDR, 7);
#else
		bit_set(TGT_PORT, TGT_TDI);
#endif
	} else {
#ifdef USE_USI_SPI
		bit_clear(USIDR, 7);
#else
		bit_clear(TGT_PORT, TGT_TDI);
#endif
	}
	if (flags & JTAG_SET_TMS) {
		bit_set(TGT_PORT, TGT_TMS);
	} else {
		bit_clear(TGT_PORT, TGT_TMS);
	}
	if (flags & JTAG_SET_TCK) {
		bit_set(TGT_PORT, TGT_TCK);
	} else {
		bit_clear(TGT_PORT, TGT_TCK);
	}
#ifdef USE_HW_SPI
	bit_clear(SPCR, SPE);
#endif
}

static inline void jtag_prepare(uint8_t flags) {
	jtag_flags = flags;
	if (flags & JTAG_CONST_TMS) {
		if (flags & JTAG_TMS_HIGH) {
			bit_set(TGT_PORT, TGT_TMS);
		} else {
			bit_clear(TGT_PORT, TGT_TMS);
		}
	}
	jtag_bytes = 0;
	jtag_bits = 0;
	jtag_bitoff = 0x80;

	if (flags & JTAG_USE_DELAY) {
		cli();
		CLKPR = (1<<CLKPCE);
		CLKPR = clkpr;
		sei();
#ifdef USE_HW_SPI
		SPSR = spsr;
		SPCR = spcr;
#endif
	} else {
		cli();
		CLKPR = (1<<CLKPCE);
#ifdef F_SCALER
		CLKPR = F_SCALER;
#else
		CLKPR = 0;
#endif
		sei();
#ifdef USE_HW_SPI
		SPSR = (1<< SPI2X);
		SPCR = (0 << SPR0) | (1 << SPE) | (1 << MSTR);
#endif
	}
}

static inline void jtag_send_8(uint8_t flags, uint8_t data, uint8_t bits) {
	uint8_t rcv = 0;
#ifdef USE_USI_SPI
	uint8_t v1,v2;

	/* F_CPU/6 bitclock(max) */
	uint8_t i;
	v1 = (1<<USIWM0)|(1<<USITC);
	v2 = (1<<USIWM0)|(1<<USITC)|(1<<USICLK);
	USIDR = data;
	for (i=0; i<8; i++) {
		USICR = v1;
		nop();
		nop();
		USICR = v2;
	}
	rcv = USIDR;
#elif defined(USE_HW_SPI)
	SPDR = data;

	while (bit_is_clear(SPSR ,SPIF))
		;
	rcv = SPDR;
#else
	uint8_t i;
	for (i=0; i<8; i++) {
		rcv <<= 1;
		if (data & 0x80) {
			bit_set(TGT_PORT, TGT_TDI);
		} else {
			bit_clear(TGT_PORT, TGT_TDI);
		}
		data <<= 1;
		bit_set(TGT_PORT, TGT_TCK);
		if (bit_is_set(TGT_PIN, TGT_TDO)) {
			rcv |= 1;
		}
		nop();
		bit_clear(TGT_PORT, TGT_TCK);
	}
#endif
	if (flags & JTAG_RECIEVE) {
		jtag_buf[jtag_bytes++] = rcv;
	}
}

static void jtag_send_d(uint8_t flags, uint8_t data, uint8_t bits) {
	uint8_t i;

#ifdef USE_HW_SPI
	if ((bits == 8) && (flags & JTAG_CONST_TMS)) {
		SPDR = data;

		while (bit_is_clear(SPSR ,SPIF))
			;
		data = SPDR;
		if (flags & JTAG_RECIEVE) {
			jtag_buf[jtag_bytes++] = data;
		}
		return;
	}
	bit_clear(TGT_PORT, TGT_TCK);
	SPCR =0;
	SPSR =0;
#endif
	for (i=0; i<bits; i++) {
		if (data & 0x80) {
#ifdef USE_USI_SPI
			bit_set(USIDR, 7);
#else
			bit_set(TGT_PORT, TGT_TDI);
#endif
		} else {
#ifdef USE_USI_SPI
			bit_clear(USIDR, 7);
#else
			bit_clear(TGT_PORT, TGT_TDI);
#endif
		}
		data <<= 1;
		if (!(flags & JTAG_CONST_TMS)) {
			if (data & 0x80) {
				bit_set(TGT_PORT, TGT_TMS);
			} else {
				bit_clear(TGT_PORT, TGT_TMS);
			}
			data <<= 1;
		}
		__delay_clk(jtag_delay);
		bit_set(TGT_PORT, TGT_TCK);
		if (flags & JTAG_RECIEVE) {
			if (bit_is_set(TGT_PIN, TGT_TDO)) {
				jtag_buf[jtag_bytes] |= jtag_bitoff;
			} else {
				jtag_buf[jtag_bytes] &= ~jtag_bitoff;
			}
			jtag_bits++;
			jtag_bitoff >>= 1;
			if (jtag_bitoff == 0) {
				jtag_bitoff = 0x80;
				jtag_bytes++;
			}
		}
		__delay_clk(jtag_delay);
		bit_clear(TGT_PORT, TGT_TCK);
	}
}

void serjtag_open() {
	extmode = 0;
}

void serjtag_close() {
	term_spi();
}

void serjtag_poll() {
    uint8_t ch;
    if (can_getc(1)) {
        ch = getc();
	if (ch == 'S') { // get Software
		putc('S');
		putc('E');
		putc('R');
		putc('J');
		putc('T');
		putc('A');
		putc('G');
	} else if (ch == 'V' || ch == 'v') { // get Version 'V':sw 'v':hw
		extmode = 0;
		putc('2'); // Revision
 #if defined(ADC_EXTENSION) && defined(FCNT_EXTENSION) && defined(OSC_EXTENSION)
		putc('3');
 #elif defined(ADC_EXTENSION) && defined(FCNT_EXTENSION)
		putc('2');
 #elif defined(ADC_EXTENSION)
		putc('1');
 #else
		putc('0');
 #endif
	} else if (extmode == 2) {
	  uint8_t bytes,bits,flags;
	  uint8_t i,r;

	  serjtag_busy_led_on();
	  if (ch == 'r') { // Request Recieved Data
		getc(); getc(); // skip header
		flags = jtag_bits & JTAG_BITS;
		bytes = jtag_bytes;
		if (flags) bytes++;
		putc('R');
		putc(flags);
		putc(jtag_bytes);
		par_putcx_start();
		for (i=0 ; i< bytes; i++) {
			par_putcx(jtag_buf[i]);
		}
		par_putcx_stop();
		jtag_bits = 0;
		jtag_bytes = 0;
	   } else if (ch == 's') { //  Set Port
		getc(); getc(); // skip header
		jtag_set(getc());
	   } else if (ch == 'd') { // Put TDI Stream
		bits = getc();
		bytes = getc();
		flags = (bits & ~JTAG_BITS) | JTAG_CONST_TMS;
		bits &= JTAG_BITS;
		jtag_prepare(flags);
		if (flags & JTAG_USE_DELAY) {
			for (i=0; i<bytes; i++) {
				jtag_send_d(flags, getc(), 8);
			}
		} else {
			for (i=0; i<bytes; i++) {
				jtag_send_8(flags, par_getcx(), 8);
			}
		}
		if (bits) {
			jtag_send_d(flags, getc(), bits);
		}
	    } else if (ch == 'D') { // Put TDI+TMS Stream
		bits = getc();
		bytes = getc();
		flags = (bits & ~JTAG_BITS);
		bits &= JTAG_BITS;
		for (i=0; i<bytes; i++) {
			jtag_send_d(flags, getc(), 4);
		}
		bits >>= 1;
		if (bits) {
			jtag_send_d(flags, getc(), bits);
		}
	    } else if (ch == 'c') { // Get TDO Stream
		bits = getc();
		bytes = getc();
		flags = (bits & ~JTAG_BITS) | JTAG_RECIEVE;
		bits &= JTAG_BITS;
		jtag_prepare(flags);
		r = (flags & JTAG_TDI_HIGH)? 0xff: 0;
		if (flags & JTAG_USE_DELAY) {
			for (i=0; i<bytes; i++) {
				jtag_send_d(flags, r, 8);
			}
		} else {
			for (i=0; i<bytes; i++) {
				jtag_send_8(flags, r, 8);
			}
		}
		if (bits) {
			jtag_send_d(flags, r, bits);
		}
	    }
	    serjtag_busy_led_off();
	} else if (extmode == 1) {
	    uint8_t bytes,i;
	    uint8_t dev;
	    int8_t r;

	    serjtag_busy_led_on();
	    r = -1; // device is not existed
	    dev = getc();
	    bytes = getc();
	    if (ch == 's' ) { // Send
		for (i=0; i<bytes && i< EXT_BUFSIZE; i++) {
		    ext_buf[i] = getc();
		}
		for (; i<bytes; i++) {
			getc();
		}
		bytes = i;
		if (dev == DEV_SYS) {
#ifdef ADC_EXTENSION
		} else if (dev == DEV_ADC) {
		    r = adc_send(bytes);
#endif
#ifdef OSC_EXTENSION
		} else if (dev == DEV_OSC) {
		    r = osc_send(bytes);
#endif
#ifdef FCNT_EXTENSION
		} else if (dev == DEV_FCNT) {
		    r = fcnt_send(bytes);
#endif
#ifndef NO_DEV_ECHO
	        } else if (dev == DEV_ECHO) {
		    r = bytes;
#endif
#ifdef I2C_EXTENSION
		} else if (DEV_I2C <= dev && dev <= 0x7f) {
		    r = i2c_send(dev, bytes);
#endif
		}
		putc(r);
	    } else if (ch == 'r') { // Recv
		 if (bytes > EXT_BUFSIZE)
			 bytes = EXT_BUFSIZE;
		if (dev == DEV_SYS) {
#ifdef ADC_EXTENSION
	        } else if (dev == DEV_ADC) {
		    r = adc_recv(bytes);
#endif
#ifdef OSC_EXTENSION
	        } else if (dev == DEV_OSC) {
		    r = osc_recv(bytes);
#endif
#ifdef FCNT_EXTENSION
	        } else if (dev == DEV_FCNT) {
		    r = fcnt_recv(bytes);
#endif
#ifndef NO_DEV_ECHO
	        } else if (dev == DEV_ECHO) {
		    r = bytes;
#endif
#ifdef I2C_EXTENSION
		} else if (DEV_I2C <= dev && dev <= 0x7f) {
		    r = i2c_recv(dev, bytes);
#endif
		}
		putc(r);
		for (i=0; i<r; i++) {
		    putc(ext_buf[i]);
	        }
	    } else {
		putc(r);
	    }
	    serjtag_busy_led_on();
	} else if (ch == 'e') { // Enter Extent mode 
		extmode = 1;
#ifdef I2C_EXTENSION
		init_i2c();
#endif
#ifdef ADC_EXTENSION
		init_adc();
#endif
#ifdef OSC_EXTENSION
		init_osc();
#endif

		putc('Y');
	} else if (ch == 'j') { // Enter Extent mode 2 (JTAG)
		extmode = 2;
		jtag_bytes = 0;
		jtag_bits = 0;
		init_spi();
		putc('Y');
	} else {
		putc('?');
	}
    }
}
